#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include "bergenUDPSocket.h"
#include "bergenSound.h"
#include "bergenServer.h"
#include "bergenListener.h"

#define NUM_TRIES 2
#define WAIT_TICKS 50

char * bergenVersion = "bergen client library, Version 0.4.1 - Compiled " __DATE__;

struct _sound
	{
	bergenSound * sound;
	struct _sound * next;
	};

// The constructor will make a UDP network connection to the snerd server     
// program, sending it a "ping" message to verify the connection. The argument
// host, if given, is the name or IP address of the machine which snerd is    
// running on; if it is not given, the value of the environment variable      
// BERGEN_SERVER is used; if this variable is not set, the connection is made 
// to the local machine. The argument port is the port number that the server 
// is listening to; if it is not given (or is 0), the value of the environment 
// variable BERGEN_PORT is used; if this variable is not set, the default port
// of 5900 is used.
bergenServer::bergenServer(char *server,int port)
	{
	soundList_ = NULL;
	path_ = NULL;
	listenerClient_ = NULL;

	if (server)
		serverName_ = strdup(server);
	else if (getenv("BERGEN_SERVER"))
		serverName_ = strdup(getenv("BERGEN_SERVER"));
	else
		serverName_ = "localhost";
	if (port)
		serverPort_ = port;
	else if (getenv("BERGEN_PORT"))
		serverPort_ = atoi(getenv("BERGEN_PORT"));
	else
		serverPort_ = 5900;
	sndSocket_ = new bergenUDPSocket(0);
	sndSocket_->setDestination(serverName_,serverPort_);
	sndSocket_->setBlocking(0);
	rcvSocket_ = new bergenUDPSocket(serverPort_+1);
	rcvSocket_->setBlocking(0);
	connect();
	listenerClient_ = new bergenListener(this);
	if (getenv("BERGEN_PATH"))
		addPath(getenv("BERGEN_PATH"));
	}


// The destructor will automatically delete all sounds which are in its list 
// of bergenSound objects (maintained by addSound() and removeSound).
bergenServer::~bergenServer(void)
	{
	struct _sound *s,*next=NULL;
	for (s=soundList_; s; s = next)
		{
		next = s->next;
		delete s->sound;
		delete s;
		}
	if (path_)
		free(path_);
	if (listenerClient_)
	 	free(listenerClient_);
	}


// Tries to establish a connection to the snerd server. Error messages if no
// reply from the snerd server or no socket is given in the bergenServer
void bergenServer::connect(void)
	{
	char replybuf[256], *msgStrings[32];
	int numStrings;
	int tries, wait;
	if (!sndSocket_ || !rcvSocket_)
		{
		fprintf(stderr,"ERROR (bergenServer::connect()): don't have a socket to use for connection\n");
		return;
		}
	for (tries=0; tries < NUM_TRIES; tries++)
		{
		sendMessage("-1 ping");
		for (wait=0; wait < WAIT_TICKS; wait++)
			{
			if (rcvSocket_->receive(replybuf,sizeof(replybuf)) > 0)
				{
				bool exit = false;
				numStrings = splitMessage(replybuf, msgStrings, 32);
				if (numStrings == 2 && !strcasecmp(msgStrings[1],"pong"))
					exit = true;
				while (numStrings)
					{
					numStrings--;
					free(msgStrings[numStrings]);
					}
				if (exit)
					break;
				}
			usleep(10000);
			}
		if (wait < WAIT_TICKS)
			break;
		}
	if (tries == NUM_TRIES)
		{
		fprintf(stderr,"WARNING: bergenServer failed to connect to server (%s)\n",serverName_);
		delete sndSocket_;
		sndSocket_ = NULL;
		delete rcvSocket_;
		rcvSocket_ = NULL;
		}
	}


// Re-initializes the connection to the server program, and re-creates all the
// sound objects (that it manages) on the server. This can be used to reset
// things if snerd crashes & is restarted, or if it is started after the
// bergenServer object is created. 
void bergenServer::reset(void)
	{
	struct _sound *s;
	for (s=soundList_; s; s = s->next)
		s->sound->killRemote();
	if (!sndSocket_)
		{
		sndSocket_ = new bergenUDPSocket(0);
		sndSocket_->setDestination(serverName_,serverPort_);
		rcvSocket_ = new bergenUDPSocket(serverPort_+1);
		}
	connect();
	if (sndSocket_)
		for (s=soundList_; s; s = s->next)
			s->sound->createRemote();
	}


// If a message was received over the socket, the message is send to 
// splitMessage for getting messages isolated. The isolated massages are
// parsed by sending them to parseMessage if they are no sound related ones.
// Sound messages are forwarded to the according sound object
void bergenServer::processMessages(void)
	{
	char message[256], *msgStrings[32];
	int numStrings;
	if (rcvSocket_)
		{
		while (rcvSocket_->receive(message,sizeof(message)) > 0)
			{
			numStrings = splitMessage(message, msgStrings, 32);
			if (numStrings <= 0)
				continue;
			int handle = atoi(msgStrings[0]);
			bergenSound* sound = findHandle(handle);
			if (sound)
				sound->message(msgStrings+1,numStrings-1);
			else
				printf("WARNING: failed to find sound with handle %d	 (message:%s)\n",handle,msgStrings[1]);
			while (numStrings)
				{
				numStrings--;
				free(msgStrings[numStrings]);
				}
			}
		}
	}


// Adds a sound to the server object's list. This is used internally, by the
// bergenSound class, and should not be called by applications.
// The list consists of the references to the sound objects. The sound object
// data is managed by snerd 
void bergenServer::addSound(bergenSound * sound)
	{
	struct _sound *s = new struct _sound;
	s->sound = sound;
	s->next = soundList_;
	soundList_ = s;
	}


// Finds a sound given the handle
bergenSound* bergenServer::findHandle(int handle)
	{
	struct _sound *s;
	if (!soundList_)
		return NULL;
	for (s=soundList_; s; s = s->next)
		{
		if (s->sound->handle() == handle)
			return s->sound;
		}
	return NULL;
	}


// Removes a sound from the server object's list. This is used internally, by 
// the bergenSound class, and should not be called by applications 
void bergenServer::removeSound(bergenSound * sound)
	{
	struct _sound *s, *removed=NULL;
	if (!soundList_)
		return;
	if (soundList_->sound == sound)
		{
		removed = soundList_;
		soundList_ = soundList_->next;
		}
	else
		for (s=soundList_; s->next; s = s->next)
			{
			if (s->next->sound == sound)
				{
				removed = s->next;
				s->next = s->next->next;
				break;
				}
			}
	if (removed)
		delete removed;
	}


// Sets the default directory path for sound objects. bergenSample objects will
// use this path if their sample file name does not include a path.
// The path is used by snerd to look up the sound files which are to be
// stored at sound server's location
void bergenServer::addPath(char *path)
	{
	if (path_)
		free(path_);
	if (path)
		{
		path_ = strdup(path);
		char cmd[1024];
		sprintf(cmd,"-1 path %s",path_);
		sendMessage(cmd);
		}
	else
		path_ = NULL;
	}


// Returns the default directory path set by setPath().
char * bergenServer::path(void)
	{
	return path_;
	}


// Sends the given text string to the server program. This is mostly for
// internal use, by the bergenSound classes. The snerd analyses the message
// and reacts on it.
void bergenServer::sendMessage(char *msg)
	{
	if (sndSocket_)
		sndSocket_->send(msg,strlen(msg)+1);
	}


// Wait for the next message from the server
int bergenServer::receiveHandle(void)
	{
	char replybuf[256], *msgStrings[32];
	int numStrings;
	int wait;
	int handle;
	if (rcvSocket_)
		{
		for (wait=0; wait < WAIT_TICKS; wait++)
			{
			if (rcvSocket_->receive(replybuf,sizeof(replybuf)) > 0)
				{
				bool exit = false;
				numStrings = splitMessage(replybuf, msgStrings, 32);
				if (numStrings == 3 && !strcasecmp(msgStrings[1],"handle"))
					{
					handle = atoi(msgStrings[2]);
					exit = true;
					}
				while (numStrings)
					{
					numStrings--;
					free(msgStrings[numStrings]);
					}
				if (exit)
					return handle;
				}
			usleep(10000);
			}
		fprintf(stderr,"WARNING: bergenServer failed to receive a handle from (%s)\n",serverName_);
		}
	return 0;
	}


// The one message string is split into isolated messages. Each message is
// either seperated by \t \n or ' '. The amount of messages found is returned
int bergenServer::splitMessage(char *message,char **strings,int max)
	{
	int i=0, start, len, count=0;
	while ((message[i]) && (count < max))
		{
		i += strspn(message+i, " \t\n");
		if (!message[i])
			break;
		start = i;
		len = strcspn(message+start, " \t\n");
		strings[count] = (char *) malloc(len+1);
		strncpy(strings[count],message+start,len);
		strings[count][len] = '\0';
		count++;
		i += len;
		}
	return count;
	}
