#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <sys/stat.h>
#include "snerdDB.h"

#define MAX_SOUNDS 32768


// Initialises the snerdDB object with initial values and allocates memory 
// for 'replay buffer' where the samples of all audio objects are mixed into
snerdDB::snerdDB(int bufsize=1024)
	{
	bufferSize_ = bufsize;
	buffer_ = (short *) malloc(bufferSize_ * sizeof(short));
	memset(buffer_, 0, bufferSize_ * sizeof(short));
	sounds_ = (ASound **) malloc(MAX_SOUNDS * sizeof(ASound *));
	memset(sounds_, 0, MAX_SOUNDS * sizeof(ASound *));
	//as many listeners as sounds because each sound may have different listener
	Listeners_ = (Listener **) malloc(MAX_SOUNDS * sizeof(Listener *));
	memset(Listeners_, 0, MAX_SOUNDS * sizeof(Listener *));
	maxSoundIndex_ = -1;
	maxListenerIndex_ = -1;
	gain_ = 1.0;
	}

// Frees memory allocated for audio object list and 'replay' buffer 
snerdDB::~snerdDB(void)
	{
	free(buffer_);
	free(sounds_);
	free(Listeners_);
	}


// Adds a directory path to the path list
void snerdDB::addPath(char * path)
	{
	if (path_[0] != 0)
		strcat(path_,":");
	strcat(path_,path);
#ifdef DEBUG
	printf("SnerdDB Adding path %s\n",path_);
#endif
	}


// Returns the path list
char * snerdDB::path(void)
	{
	return path_;
	}


// Finds a file in the path list
int snerdDB::findFile(char* filename, char * fullpath)
	{
	struct stat statbuf;
	if (!filename)
		{
		fullpath[0] = 0;
		return 0;
		}
	else if ((filename[0] == '/') || ((filename[0] == '.') && (filename[1] == '/')))
		{
		strcpy(fullpath,filename);
		if (stat(fullpath,&statbuf) == 0)
			return 1;
		else
			return 0;
		}
	else
		{
		strcpy(fullpath,filename);
		if (stat(filename,&statbuf) == 0)
			{
			return 1;
			}
		if (path_[0] != 0)
			{
			char *tokenChar;
			char path[256];
			strcpy(path,path_);
			tokenChar = strtok(path,":");
			do
				{
				sprintf(fullpath,"%s/%s",tokenChar,filename);
				if (stat(fullpath,&statbuf) == 0)
					return 1;
				}
			while (tokenChar = strtok(NULL,":"));
			strcpy(path,path_);
			tokenChar = strtok(path,":");
			sprintf(fullpath,"%s/%s",tokenChar,filename);
			}
		}
	return 0;
	}


// Adds a sound object to the list of sound objects
void snerdDB::addSound(ASound * sound)
	{
	int i;
#ifdef DEBUG
	printf("SnerdDB Adding sound %d\n",sound->handle());
#endif
	for (i=0; i < MAX_SOUNDS; i++)
		if (!sounds_[i])
			{
			sounds_[i] = sound;
			if (i > maxSoundIndex_)
				maxSoundIndex_ = i;
			return;
			}
	fprintf(stderr,"ERROR: snerdDB::addSound() - list is full; cannot add new sound\n");
	}


// Adds a listener object to the list of listener objects 
void snerdDB::addListener(Listener * listener)
	{
	int i;
#ifdef DEBUG
	printf("Adding listener %d\n",listener->handle());
#endif
	for (i=0; i < MAX_SOUNDS; i++)
		if (!Listeners_[i])
			{
			Listeners_[i] = listener;
			if (i > maxListenerIndex_)
				maxListenerIndex_ = i;
			return;
			}
	fprintf(stderr,"ERROR: snerdDB::addListener() - list is full; cannot add new listener\n");
	}


// Searches the handle to a sound object by given id and returns the
// reference 
ASound * snerdDB::findHandle(int h)
	{
	int i;
	for (i=0; i <= maxSoundIndex_; i++)
		if (sounds_[i])
			{
			if (h == sounds_[i]->handle())
				return sounds_[i];
			}
	return NULL;
	}


// Searches the handle to a listener object by given id and returns the 
Listener * snerdDB::findListener(int h)
	{
	#ifdef DEBUG
		printf("snerdDB FindListener\n");
	#endif	
	int i;
	#ifdef DEBUG
		printf("SnerdDB maxlistenerIndex is\n");
		printf("%d",maxSoundIndex_);
	#endif	
	for (i=0; i <= maxListenerIndex_; i++)
		if (Listeners_[i])
			{
			if (h == Listeners_[i]->handle())
				return Listeners_[i];
			}
	return NULL;
	}


// Calls every sound objects update routine and mixes the updated buffers
// down to one the 'replay' buffer
void snerdDB::update(void)
	{
	#ifdef DEBUG
		printf("DB update\n");
	#endif
	int i, j;
	memset(buffer_, 0, bufferSize_ * sizeof(short));
	#ifdef DEBUG
		printf("maxsoundindex %d\n",maxSoundIndex_);
	#endif
	for (i=0; i <= maxSoundIndex_; i++)
		{
		//delete sounds that were remotelete marked for delete
		if ((sounds_[i]) && (sounds_[i]->removable()))
			{
			delete sounds_[i];
			sounds_[i] = NULL;
			if (i == maxSoundIndex_)
				{
				while ((maxSoundIndex_ >= 0) && (sounds_[maxSoundIndex_] == NULL))
					maxSoundIndex_--;
				}
			}
		//delete listeners that were remotelete marked for delete
		if ((Listeners_[i]) && (Listeners_[i]->removable()))
			{
			delete Listeners_[i];
			Listeners_[i] = NULL;
			if (i == maxListenerIndex_)
				{
				while ((maxListenerIndex_ >= 0) && (Listeners_[maxListenerIndex_] == NULL))
					maxListenerIndex_--;
				}
			}
		}
	for (i=0; i <= maxListenerIndex_; i++)
		{
		if (Listeners_[i])
			Listeners_[i]->updatePosition();
		}
	for (i=0; i <= maxSoundIndex_; i++)
		{
		if ((sounds_[i]) && (sounds_[i]->active()))
			{
			short * sbuf = sounds_[i]->updateBuffer();
			if (sbuf)
				for (j=0; j < bufferSize_; j++)
					buffer_[j] += sbuf[j];
			}
		}
	if (gain_ != 1.0f)
		{
		for (i=0; i < bufferSize_; i++)
			buffer_[i] = (short)(buffer_[i] * gain_);
		}
	}


// Deletes all Sounds in snerdDB
void snerdDB::reset(void)
	{
	int i;
	for (i=0; i <= maxSoundIndex_; i++)
		{
		if (sounds_[i])
			{
			sounds_[i]->kill();
			delete sounds_[i];
			}
		}
	for (i=0; i <= maxListenerIndex_; i++)
		{
		if (Listeners_[i])
			{
			delete Listeners_[i];
			}
		}
	memset(sounds_, 0, MAX_SOUNDS * sizeof(ASound *));
	maxSoundIndex_ = -1;
	memset(Listeners_, 0, MAX_SOUNDS * sizeof(Listener *));
	maxListenerIndex_ = -1;
	}


// Sets the gain for the snerdDB 
void snerdDB::setGain(float g)
	{
	gain_ = g;
	}
